import dash_bootstrap_components as dbc

# See https://dash-bootstrap-components.opensource.faculty.ai/docs/
# for using dash_bootstrap_components for simplifying programmatic layout.
# Set this to an empty list if you are not using any external sheets
external_stylesheets = [dbc.themes.YETI]
